@@ -226,7 +226,7 @@ get_parms (struct dump_file *file)
   return ferror (file->fp) ? GDBM_FILE_READ_ERROR : 0;
 }
 
-int
+static int
 get_len (const char *param, size_t *plen)
 {
   unsigned long n;
@@ -247,7 +247,7 @@ get_len (const char *param, size_t *plen)
   return GDBM_ILLEGAL_DATA;
 }
 
-int
+static int
 read_record (struct dump_file *file, char *param, int n, datum *dat)
 {
   int rc;
@@ -391,12 +391,24 @@ _set_gdbm_meta_info (GDBM_FILE dbf, char *param, int meta_mask)
 }
 
 int
+_gdbm_str2fmt (char const *str)
+{
+  if (strcmp (str, "numsync") == 0)
+    return GDBM_NUMSYNC;
+  if (strcmp (str, "standard") == 0)
+    return 0;
+  return -1;
+}
+
+static int
 _gdbm_load_file (struct dump_file *file, GDBM_FILE dbf, GDBM_FILE *ofp,
 		 int replace, int meta_mask)
 {
   char *param = NULL;
   int rc;
   GDBM_FILE tmp = NULL;
+  int format = 0;
+  const char *p;
   
   rc = get_parms (file);
   if (rc)
@@ -411,17 +423,42 @@ _gdbm_load_file (struct dump_file *file, GDBM_FILE dbf, GDBM_FILE *ofp,
   else
     return GDBM_ILLEGAL_DATA;
 
+  if ((p = getparm (file->header, "format")) != NULL)
+    {
+      int n = _gdbm_str2fmt (p);
+      if (n != -1)
+	format = n;
+      /* FIXME: other values silently ignored */
+    }
+      
   if (!dbf)
     {
+      int flags = replace ? GDBM_WRCREAT : GDBM_NEWDB;
       const char *filename = getparm (file->header, "file");
+      
       if (!filename)
 	return GDBM_NO_DBNAME;
-      tmp = gdbm_open (filename, 0,
-		       replace ? GDBM_WRCREAT : GDBM_NEWDB, 0600, NULL);
+
+      tmp = gdbm_open (filename, 0, flags | format, 0600, NULL);
       if (!tmp)
 	return gdbm_errno;
       dbf = tmp;
     }
+
+  if (format)
+    {
+      /*
+       * If the database is already in the requested format, the call to
+       * gdbm_convert will return 0 immediately.
+       */
+      if (gdbm_convert (dbf, format))
+	{	
+	  rc = gdbm_errno;
+	  if (tmp)
+	    gdbm_close (tmp);
+	  return rc;
+	}
+    }	  
   
   param = file->header;
   while (1)
@@ -529,7 +566,7 @@ xdatum_read (FILE *fp, datum *d, size_t *pdmax)
   return c;
 }
 
-int
+static int
 gdbm_load_bdb_dump (struct dump_file *file, GDBM_FILE dbf, int replace)
 {
   datum xd[2];
