@@ -22,15 +22,6 @@
 #include "gdbmdefs.h"
 #include <stddef.h>
 
-/* Determine our native magic number and bail if we can't. */
-#if SIZEOF_OFF_T == 4
-# define GDBM_MAGIC	GDBM_MAGIC32
-#elif SIZEOF_OFF_T == 8
-# define GDBM_MAGIC	GDBM_MAGIC64
-#else
-# error "Unsupported off_t size, contact GDBM maintainer.  What crazy system is this?!?"
-#endif
-
 static void
 compute_directory_size (blksize_t block_size,
 			int *ret_dir_size, int *ret_dir_bits)
@@ -57,39 +48,40 @@ bucket_element_count (size_t bucket_size)
   return (bucket_size - sizeof (hash_bucket)) / sizeof (bucket_element) + 1;
 }
 
-static int
-validate_header (gdbm_file_header const *hdr, struct stat const *st)
+static void
+gdbm_header_avail (gdbm_file_header *hdr,
+		   avail_block **avail_ptr, size_t *avail_size,
+		   gdbm_ext_header **exhdr)
 {
-  int dir_size, dir_bits;
-  int result = GDBM_NO_ERROR;
-  
-  /* Is the magic number good? */
-  if (hdr->header_magic != GDBM_MAGIC)
-    {
-      switch (hdr->header_magic)
-	{
-	case GDBM_OMAGIC:
-	  /* OK */
-	  break;
-
-	case GDBM_OMAGIC_SWAP:
-	case GDBM_MAGIC32_SWAP:
-	case GDBM_MAGIC64_SWAP:
-	  return GDBM_BYTE_SWAPPED;
-
-	case GDBM_MAGIC32:
-	case GDBM_MAGIC64:
-	  return GDBM_BAD_FILE_OFFSET;
-
-	default:
-	  return GDBM_BAD_MAGIC_NUMBER;
-	}
+  switch (hdr->header_magic)
+    { 
+    case GDBM_OMAGIC:
+    case GDBM_MAGIC: 
+      *exhdr = NULL;
+      *avail_ptr = &((gdbm_file_standard_header*)hdr)->avail;
+      *avail_size = (hdr->block_size -
+		     offsetof (gdbm_file_standard_header, avail));
+      break;
+      
+    case GDBM_NUMSYNC_MAGIC:
+      *exhdr = &((gdbm_file_extended_header*)hdr)->ext;
+      *avail_ptr = &((gdbm_file_extended_header*)hdr)->avail;
+      *avail_size = (hdr->block_size -
+		     offsetof (gdbm_file_extended_header, avail));
+      break;
     }
+}
+
+static int
+validate_header_std (gdbm_file_header const *hdr, struct stat const *st)
+{
+  int result = GDBM_NO_ERROR;
+  int dir_size, dir_bits;
   
   if (!(hdr->block_size > 0
 	&& hdr->block_size > sizeof (gdbm_file_header)
 	&& hdr->block_size - sizeof (gdbm_file_header) >=
-	sizeof(hdr->avail.av_table[0])))
+	sizeof(avail_block)))
     {
       return GDBM_BLOCK_SIZE_ERROR;
     }
@@ -117,19 +109,96 @@ validate_header (gdbm_file_header const *hdr, struct stat const *st)
   if (hdr->dir_bits != dir_bits)
     return GDBM_BAD_HEADER;
   
-  if (!(hdr->bucket_size > sizeof(hash_bucket)))
+  if (!(hdr->bucket_size > sizeof (hash_bucket)))
     return GDBM_BAD_HEADER;
 
   if (hdr->bucket_elems != bucket_element_count (hdr->bucket_size))
     return GDBM_BAD_HEADER;
 
-  if (((hdr->block_size - sizeof (gdbm_file_header)) / sizeof(avail_elem) + 1)
-      != hdr->avail.size)
+  return result;
+}
+
+static int
+validate_header_numsync (gdbm_file_header const *hdr, struct stat const *st)
+{
+  int result = GDBM_NO_ERROR;
+  int dir_size, dir_bits;
+  
+  if (!(hdr->block_size > 0
+	&& hdr->block_size > (sizeof (gdbm_file_header) + sizeof (gdbm_ext_header))
+	&& hdr->block_size - (sizeof (gdbm_file_header) + sizeof (gdbm_ext_header)) >=
+	sizeof(avail_block)))
+    {
+      return GDBM_BLOCK_SIZE_ERROR;
+    }
+
+  /* Technically speaking, the condition below should read
+         hdr->next_block != st->st_size
+     However, gdbm versions prior to commit 4e819c98 could leave
+     hdr->next_block pointing beyond current end of file. To ensure
+     backward compatibility with these versions, the condition has been
+     slackened to this: */
+  if (hdr->next_block < st->st_size)
+    result = GDBM_NEED_RECOVERY;
+
+  /* Make sure dir and dir + dir_size fall within the file boundary */
+  if (!(hdr->dir > 0
+	&& hdr->dir < st->st_size
+	&& hdr->dir_size > 0
+	&& hdr->dir + hdr->dir_size < st->st_size))
+    return GDBM_BAD_HEADER;
+
+  compute_directory_size (hdr->block_size, &dir_size, &dir_bits);
+  if (!(hdr->dir_size >= dir_size))
+    return GDBM_BAD_HEADER;
+  compute_directory_size (hdr->dir_size, &dir_size, &dir_bits);
+  if (hdr->dir_bits != dir_bits)
+    return GDBM_BAD_HEADER;
+  
+  if (!(hdr->bucket_size > sizeof (hash_bucket)))
+    return GDBM_BAD_HEADER;
+
+  if (hdr->bucket_elems != bucket_element_count (hdr->bucket_size))
     return GDBM_BAD_HEADER;
 
   return result;
 }
 
+static int
+validate_header (gdbm_file_header const *hdr, struct stat const *st)
+{
+  /* Is the magic number good? */
+  switch (hdr->header_magic)
+    {
+    case GDBM_OMAGIC:
+    case GDBM_MAGIC:
+      return validate_header_std (hdr, st);
+      
+    case GDBM_NUMSYNC_MAGIC:
+      return validate_header_numsync (hdr, st);
+
+    default:
+      switch (hdr->header_magic)
+	{
+	case GDBM_OMAGIC_SWAP:
+	case GDBM_MAGIC32_SWAP:
+	case GDBM_MAGIC64_SWAP:
+	case GDBM_NUMSYNC_MAGIC32_SWAP:
+	case GDBM_NUMSYNC_MAGIC64_SWAP:
+	  return GDBM_BYTE_SWAPPED;
+
+	case GDBM_MAGIC32:
+	case GDBM_MAGIC64:
+	case GDBM_NUMSYNC_MAGIC32:
+	case GDBM_NUMSYNC_MAGIC64:
+	  return GDBM_BAD_FILE_OFFSET;
+
+	default:
+	  return GDBM_BAD_MAGIC_NUMBER;
+	}
+    }
+}
+
 int
 _gdbm_validate_header (GDBM_FILE dbf)
 {
@@ -142,8 +211,7 @@ _gdbm_validate_header (GDBM_FILE dbf)
   rc = validate_header (dbf->header, &file_stat);
   if (rc == 0)
     {
-      if (gdbm_avail_block_validate (dbf, &dbf->header->avail,
-				     GDBM_HEADER_AVAIL_SIZE (dbf)))
+      if (gdbm_avail_block_validate (dbf, dbf->avail, dbf->avail_size))
 	rc = GDBM_BAD_AVAIL;
     }
   return rc;
@@ -356,8 +424,16 @@ gdbm_fd_open (int fd, const char *file_name, int block_size,
 	}
 
       /* Set the magic number and the block_size. */
-      dbf->header->header_magic = GDBM_MAGIC;
+      if (flags & GDBM_NUMSYNC)
+	dbf->header->header_magic = GDBM_NUMSYNC_MAGIC;
+      else
+	dbf->header->header_magic = GDBM_MAGIC;
+
+      /*
+       * Set block size.  It must be initialized for gdbm_header_avail to work.
+       */
       dbf->header->block_size = block_size;
+      gdbm_header_avail (dbf->header, &dbf->avail, &dbf->avail_size, &dbf->xheader);
       dbf->header->dir_size = dir_size;
       dbf->header->dir_bits = dir_bits;
 
@@ -395,11 +471,11 @@ gdbm_fd_open (int fd, const char *file_name, int block_size,
 	dbf->dir[index] = 2*dbf->header->block_size;
 
       /* Initialize the active avail block. */
-      dbf->header->avail.size
-	= ( (dbf->header->block_size - sizeof (gdbm_file_header))
-	 / sizeof (avail_elem)) + 1;
-      dbf->header->avail.count = 0;
-      dbf->header->avail.next_block = 0;
+      dbf->avail->size = (dbf->avail_size - offsetof(avail_block, av_table))
+	                  / sizeof (avail_elem);
+      dbf->avail->count = 0;
+      dbf->avail->next_block = 0;
+      
       dbf->header->next_block  = 4*dbf->header->block_size;
 
       /* Write initial configuration to the file. */
@@ -502,7 +578,7 @@ gdbm_fd_open (int fd, const char *file_name, int block_size,
 	}
       
       memcpy (dbf->header, &partial_header, sizeof (partial_header));
-      if (_gdbm_full_read (dbf, &dbf->header->avail.av_table[1],
+      if (_gdbm_full_read (dbf, dbf->header + 1,
 			   dbf->header->block_size - sizeof (gdbm_file_header)))
 	{
 	  if (!(flags & GDBM_CLOERROR))
@@ -510,9 +586,19 @@ gdbm_fd_open (int fd, const char *file_name, int block_size,
 	  SAVE_ERRNO (gdbm_close (dbf));
 	  return NULL;
 	}
+      gdbm_header_avail (dbf->header, &dbf->avail, &dbf->avail_size, &dbf->xheader);
 
-      if (gdbm_avail_block_validate (dbf, &dbf->header->avail,
-				     GDBM_HEADER_AVAIL_SIZE (dbf)))
+      if (((dbf->header->block_size -
+	    (GDBM_HEADER_AVAIL_OFFSET (dbf) +
+	     sizeof (avail_block))) / sizeof (avail_elem) + 1) != dbf->avail->size)
+	{
+	  if (!(flags & GDBM_CLOERROR))
+	    dbf->desc = -1;
+	  GDBM_SET_ERRNO2 (NULL, GDBM_BAD_HEADER, FALSE, GDBM_DEBUG_OPEN);
+	  return NULL;
+	}
+      
+      if (gdbm_avail_block_validate (dbf, dbf->avail, dbf->avail_size))
 	{
 	  if (!(flags & GDBM_CLOERROR))
 	    dbf->desc = -1;
@@ -674,3 +760,163 @@ _gdbm_file_size (GDBM_FILE dbf, off_t *psize)
   *psize = dbf->file_size;
   return 0;
 }
+
+/*
+ * Convert from numsync to the standard GDBM format.  This is pretty
+ * straightforward: the avail block gets expanded by
+ * sizeof(gdbm_ext_header), so we only need to shift the avail block
+ * up this number of bytes, change the avail and avail_size pointers
+ * and update the magic number.
+ */
+static int
+_gdbm_convert_from_numsync (GDBM_FILE dbf)
+{
+  avail_block *old_avail = dbf->avail;
+
+  /* Change the magic number */
+  dbf->header->header_magic = GDBM_MAGIC;
+  /* Update avail pointer and size */
+  gdbm_header_avail (dbf->header, &dbf->avail, &dbf->avail_size, &dbf->xheader);
+  
+  /* Move data up */
+  memmove (dbf->avail, old_avail, dbf->avail_size - sizeof (gdbm_ext_header));
+
+  /* Fix up the avail table size */
+  dbf->avail->size = (dbf->avail_size - offsetof (avail_block, av_table))
+	                  / sizeof (avail_elem);
+
+  dbf->header_changed = TRUE;
+  
+  return 0;
+}
+
+/*
+ * Convert the database from the standard to extended (numsync) format.
+ * The avail block shrinks by sizeof(gdbm_ext_header) bytes.  The av_table
+ * entries that don't fit into the new size need to be returned to the
+ * avail pool using the _gdbm_free call.
+ */
+static int
+_gdbm_convert_to_numsync (GDBM_FILE dbf)
+{
+  avail_block *old_avail = dbf->avail;
+  size_t old_avail_size = dbf->avail->size;
+  size_t n; /* Number of elements to return to the pool */
+  int rc;
+  avail_elem *av = NULL;
+  
+  /* Change the magic number */
+  dbf->header->header_magic = GDBM_NUMSYNC_MAGIC;
+  /* Update avail pointer and size */
+  gdbm_header_avail (dbf->header, &dbf->avail, &dbf->avail_size, &dbf->xheader);
+  /*
+   * Compute new av_table size.
+   * NOTE: Don't try to modify dbf->avail until the final move, otherwise
+   * the available block would end up clobbered.  All modifications are
+   * applied to old_avail.
+   */
+  old_avail->size = (dbf->avail_size - offsetof (avail_block, av_table))
+	                  / sizeof (avail_elem);
+  /* Compute the number of avail elements that don't fit in the new table. */
+  n = old_avail_size - old_avail->size;
+  if (n > 0)
+    {
+      /* Stash them away */
+      av = calloc (n, sizeof (av[0]));
+      if (!av)
+	{
+	  GDBM_SET_ERRNO (dbf, GDBM_MALLOC_ERROR, FALSE);
+	  return -1;
+	}
+      n = 0;
+      while (old_avail->count > old_avail->size)
+	{
+	  old_avail->count--;
+	  av[n++] = old_avail->av_table[old_avail->count];
+	}
+    }
+
+  /*
+   * Move the modified avail block into its new place.  From now on,
+   * old_avail may not be use.  The database header is in consistent
+   * state and all modifications should be applied to it directly.
+   */
+  memmove (dbf->avail, old_avail, dbf->avail_size);
+
+  /* Initialize the extended header */
+  memset (dbf->xheader, 0, sizeof (dbf->xheader[0]));
+
+  rc = 0; /* Assume success */
+  
+  if (av)
+    {
+      /* Return stashed av_table elements to the available pool. */
+      /* _gdbm_free needs a non-NULL bucket, so get one: */
+      if (!dbf->bucket)
+	rc = _gdbm_get_bucket (dbf, 0);
+      if (rc == 0)
+	{
+	  size_t i;
+	  
+	  for (i = 0; i < n; i++)
+	    {
+	      rc = _gdbm_free (dbf, av[i].av_adr, av[i].av_size);
+	      if (rc)
+		break;
+	    }
+	}
+      free (av);
+    }
+
+  dbf->header_changed = TRUE;
+  
+  return rc;
+}
+
+int
+gdbm_convert (GDBM_FILE dbf, int flag)
+{
+  int rc;
+  
+  /* Return immediately if the database needs recovery */	
+  GDBM_ASSERT_CONSISTENCY (dbf, -1);
+  
+  /* First check to make sure this guy is a writer. */
+  if (dbf->read_write == GDBM_READER)
+    {
+      GDBM_SET_ERRNO2 (dbf, GDBM_READER_CANT_STORE, FALSE,
+		       GDBM_DEBUG_STORE);
+      return -1;
+    }
+
+  switch (flag)
+    {
+    case 0:
+    case GDBM_NUMSYNC:
+      break;
+
+    default:
+      GDBM_SET_ERRNO2 (dbf, GDBM_ILLEGAL_DATA, FALSE,
+		       GDBM_DEBUG_STORE);
+      return -1;
+    }
+
+  rc = 0;
+  switch (dbf->header->header_magic)
+    {
+    case GDBM_OMAGIC:
+    case GDBM_MAGIC:
+      if (flag == GDBM_NUMSYNC)
+	rc = _gdbm_convert_to_numsync (dbf);
+      break;
+      
+    case GDBM_NUMSYNC_MAGIC:
+      if (flag == 0)
+	rc = _gdbm_convert_from_numsync (dbf);	
+    }
+
+  if (rc == 0)
+    rc = _gdbm_end_update (dbf);
+  
+  return 0;
+}
