@@ -23,18 +23,20 @@
 #include <signal.h>
 #include <pwd.h>
 #include <sys/ioctl.h>
+#include <sys/wait.h>
 #include <termios.h>
 #include <stdarg.h>
 #ifdef HAVE_LOCALE_H
 # include <locale.h>
 #endif
 
-char *file_name = NULL;       /* Database file name */   
+char *file_name = NULL;       /* Database file name */
+int file_descr = -1;          /* Database file descriptor */
 GDBM_FILE gdbm_file = NULL;   /* Database to operate upon */
 datum key_data;               /* Current key */
 datum return_data;            /* Current data */
 int open_mode;                /* Default open mode */
-
+int open_format;              /* Default format for the open command */
 unsigned input_line;
 
 
@@ -47,15 +49,16 @@ closedb (void)
       gdbm_file = NULL;
       free (file_name);
       file_name = NULL;
+      file_descr = -1;
     }
 }
 
 static int
-opendb (char *dbname)
+opendb (char *dbname, int fd)
 {
   int cache_size = 0;
   int block_size = 0;
-  int flags = 0;
+  int flags = open_format;
   int filemode;
   GDBM_FILE db;
   
@@ -96,7 +99,10 @@ opendb (char *dbname)
   if (variable_get ("filemode", VART_INT, (void**) &filemode))
     abort ();
 
-  db = gdbm_open (dbname, block_size, open_mode | flags, filemode, NULL);
+  if (fd > 0)
+    db = gdbm_fd_open (fd, dbname, block_size, open_mode | flags, NULL);
+  else
+    db = gdbm_open (dbname, block_size, open_mode | flags, filemode, NULL);
 
   if (db == NULL)
     {
@@ -137,10 +143,9 @@ checkdb (void)
       if (!file_name)
 	{
 	  file_name = estrdup (GDBMTOOL_DEFFILE);
-	  terror (_("warning: using default database file %s"),
-			file_name);
+	  terror (_("warning: using default database file %s"),	file_name);
 	}
-      return opendb (file_name);
+      return opendb (file_name, file_descr);
     }
   return 0;
 }
@@ -256,8 +261,14 @@ static int
 avail_list_print (avail_block *avblk, off_t n, void *data)
 {
   FILE *fp = data;
-  fprintf (fp, _("\nblock = %lu\nsize  = %d\ncount = %d\n"),
-	   (unsigned long) n, avblk->size, avblk->count);
+  
+  fputc ('\n', fp);
+  if (n == 0)//FIXME
+    fprintf (fp, "%s", _("header block"));
+  else
+    fprintf (fp, _("block = %lu"), (unsigned long) n);
+  fprintf (fp, _("\nsize  = %d\ncount = %d\n"),
+	   avblk->size, avblk->count);
   av_table_display (avblk->av_table, avblk->count, fp);
   return 0;
 }
@@ -265,10 +276,6 @@ avail_list_print (avail_block *avblk, off_t n, void *data)
 void
 _gdbm_print_avail_list (FILE *fp, GDBM_FILE dbf)
 {
-  /* Print the the header avail block.  */
-  fprintf (fp, _("\nheader block\nsize  = %d\ncount = %d\n"),
-	   dbf->header->avail.size, dbf->header->avail.count);
-  av_table_display (dbf->header->avail.av_table, dbf->header->avail.count, fp);
   if (gdbm_avail_traverse (dbf, avail_list_print, fp))
     terror ("%s", gdbm_strerror (gdbm_errno));
 }
@@ -341,7 +348,7 @@ open_handler (struct handler_param *param)
 {
   char *name = tildexpand (PARAM_STRING (param, 0));
   closedb ();
-  if (opendb (name) == 0)
+  if (opendb (name, -1) == 0)
     file_name = name;
   else
     free (name);
@@ -783,8 +790,28 @@ void
 print_header_handler (struct handler_param *param)
 {
   FILE *fp = param->fp;
+  char const *type;
+
+  switch (gdbm_file->header->header_magic)
+    {
+    case GDBM_OMAGIC:
+      type = "GDBM (old)";
+      break;
+
+    case GDBM_MAGIC:
+      type = "GDBM (standard)";
+      break;
+
+    case GDBM_NUMSYNC_MAGIC:
+      type = "GDBM (numsync)";
+      break;
+
+    default:
+      abort ();
+    }
   
   fprintf (fp, _("\nFile Header: \n\n"));
+  fprintf (fp, _("  type         = %s\n"), type);
   fprintf (fp, _("  table        = %lu\n"),
 	   (unsigned long) gdbm_file->header->dir);
   fprintf (fp, _("  table size   = %d\n"), gdbm_file->header->dir_size);
@@ -795,11 +822,249 @@ print_header_handler (struct handler_param *param)
   fprintf (fp, _("  header magic = %x\n"), gdbm_file->header->header_magic);
   fprintf (fp, _("  next block   = %lu\n"),
 	   (unsigned long) gdbm_file->header->next_block);
-  fprintf (fp, _("  avail size   = %d\n"), gdbm_file->header->avail.size);
-  fprintf (fp, _("  avail count  = %d\n"), gdbm_file->header->avail.count);
+
+  fprintf (fp, _("  avail size   = %d\n"), gdbm_file->avail->size);
+  fprintf (fp, _("  avail count  = %d\n"), gdbm_file->avail->count);
   fprintf (fp, _("  avail nx blk = %lu\n"),
-	   (unsigned long) gdbm_file->header->avail.next_block);
-}  
+	   (unsigned long) gdbm_file->avail->next_block);
+
+  if (gdbm_file->xheader)
+    {
+      fprintf (fp, _("\nExtended Header: \n\n"));
+      fprintf (fp, _("       version = %d\n"), gdbm_file->xheader->version);  
+      fprintf (fp, _("       numsync = %u\n"), gdbm_file->xheader->numsync);
+    }
+}
+
+static void
+sync_handler (struct handler_param *param)
+{
+  if (gdbm_sync (gdbm_file))
+    terror ("%s", gdbm_db_strerror (gdbm_file));
+}
+
+static void
+upgrade_handler (struct handler_param *param)
+{
+  if (gdbm_convert (gdbm_file, GDBM_NUMSYNC))
+    terror ("%s", gdbm_db_strerror (gdbm_file));
+}
+
+static void
+downgrade_handler (struct handler_param *param)
+{
+  if (gdbm_convert (gdbm_file, 0))
+    terror ("%s", gdbm_db_strerror (gdbm_file));
+}
+
+struct snapshot_status_info
+{
+  char const *code;
+  char const *descr;
+  void (*fn) (FILE *, char const *, char const *);
+};
+
+#define MODBUFSIZE 10
+
+static char *
+decode_mode (mode_t mode, char *buf)
+{
+  char *s = buf;
+  *s++ = mode & S_IRUSR ? 'r' : '-';
+  *s++ = mode & S_IWUSR ? 'w' : '-';
+  *s++ = (mode & S_ISUID
+	       ? (mode & S_IXUSR ? 's' : 'S')
+	       : (mode & S_IXUSR ? 'x' : '-'));
+  *s++ = mode & S_IRGRP ? 'r' : '-';
+  *s++ = mode & S_IWGRP ? 'w' : '-';
+  *s++ = (mode & S_ISGID
+	       ? (mode & S_IXGRP ? 's' : 'S')
+	       : (mode & S_IXGRP ? 'x' : '-'));
+  *s++ = mode & S_IROTH ? 'r' : '-';
+  *s++ = mode & S_IWOTH ? 'w' : '-';
+  *s++ = (mode & S_ISVTX
+	       ? (mode & S_IXOTH ? 't' : 'T')
+	       : (mode & S_IXOTH ? 'x' : '-'));
+  *s = '\0';
+  return buf;
+}
+
+struct error_entry
+{
+  const char *msg;
+  int gdbm_err;
+  int sys_err;
+};
+
+static void
+error_push (struct error_entry *stk, int *tos, int maxstk, char const *text,
+	    int gdbm_err, int sys_err)
+{
+  if (*tos == maxstk)
+    abort ();
+  stk += *tos;
+  ++ *tos;
+  stk->msg = text;
+  stk->gdbm_err = gdbm_err;
+  stk->sys_err = sys_err;
+}
+
+static void
+print_snapshot (char const *snapname, FILE *fp)
+{
+  struct stat st;
+  char buf[MODBUFSIZE];
+
+  if (stat (snapname, &st) == 0)
+    {
+# define MAXERRS 4
+      struct error_entry errs[MAXERRS];
+      int errn = 0;
+      int i;
+      
+      switch (st.st_mode & ~S_IFREG)
+	{
+	case S_IRUSR:
+	case S_IWUSR:
+	  break;
+	  
+	default:
+	  error_push (errs, &errn, ARRAY_SIZE (errs), N_("bad file mode"), 
+                      0, 0);
+	}
+      
+      fprintf (fp, "%s: ", snapname);
+      fprintf (fp, "%03o %s ", st.st_mode & 0777,
+	       decode_mode (st.st_mode, buf));
+      fprintf (fp, "%ld.%09ld", st.st_mtim.tv_sec, st.st_mtim.tv_nsec);
+      if (S_ISREG (st.st_mode))
+	{
+	  GDBM_FILE dbf;
+
+	  dbf = gdbm_open (snapname, 0, GDBM_READER, 0, NULL);
+	  if (dbf)
+	    {
+	      if (dbf->xheader)
+		fprintf (fp, " %u", dbf->xheader->numsync);
+	      else
+		/* TRANSLATORS: Stands for "Not Available". */
+		fprintf (fp, " %s", _("N/A"));
+	    }
+	  else if (gdbm_check_syserr (gdbm_errno))
+	    {
+	      if (errno == EACCES)
+		fprintf (fp, " ?");
+	      else
+		error_push (errs, &errn, ARRAY_SIZE (errs),
+			    N_("can't open database"),
+			    gdbm_errno, errno);
+	    }
+	  else
+	    error_push (errs, &errn, ARRAY_SIZE (errs),
+			N_("can't open database"),
+			gdbm_errno, 0);
+	}
+      else
+	error_push (errs, &errn, ARRAY_SIZE (errs),
+		    N_("not a regular file"),
+		    0, 0);
+      fputc ('\n', fp);
+      for (i = 0; i < errn; i++)
+	{
+	  fprintf (fp, "%s: %s: %s", snapname, _("ERROR"), gettext (errs[i].msg));
+	  if (errs[i].gdbm_err)
+	    fprintf (fp, ": %s", gdbm_strerror (errs[i].gdbm_err));
+	  if (errs[i].sys_err)
+	    fprintf (fp, ": %s", strerror (errs[i].sys_err));
+	  fputc ('\n', fp);
+	}	  
+    }
+  else
+    {
+      fprintf (fp, _("%s: ERROR: can't stat: %s"), snapname, strerror (errno));
+      return;
+    }
+}
+
+static void
+snapshot_print_fn (FILE *fp, char const *sa, char const *sb)
+{
+  print_snapshot (sa, fp);
+  print_snapshot (sb, fp);
+}
+
+static void
+snapshot_err_fn (FILE *fp, char const *sa, char const *sb)
+{
+  switch (errno)
+    {
+    default:
+      print_snapshot (sa, fp);
+      print_snapshot (sb, fp);
+      break;
+      
+    case EINVAL:
+      fprintf (fp, "%s.\n",
+	       _("Invalid arguments in call to gdbm_latest_snapshot"));
+      break;
+      
+    case ENOSYS:
+      fprintf (fp, "%s.\n",
+	       _("Function is not implemented: GDBM is built without crash-tolerance support"));
+      break;
+    }      
+}
+
+static struct snapshot_status_info snapshot_status_info[] = {
+  [GDBM_SNAPSHOT_OK] = {
+    "GDBM_SNAPSHOT_OK",
+    N_("Selected the most recent snapshot")
+  },
+  [GDBM_SNAPSHOT_BAD] = {
+    "GDBM_SNAPSHOT_BAD",
+    N_("Neither snapshot is readable"),
+    snapshot_print_fn
+  },
+  [GDBM_SNAPSHOT_ERR] = {
+    "GDBM_SNAPSHOT_ERR",
+    N_("Error selecting snapshot"),
+    snapshot_err_fn
+  },
+  [GDBM_SNAPSHOT_SAME] = {
+    "GDBM_SNAPSHOT_SAME",
+    N_("Snapshot modes and dates are the same"),
+    snapshot_print_fn
+  },
+  [GDBM_SNAPSHOT_SUSPICIOUS] = {
+    "GDBM_SNAPSHOT_SUSPICIOUS",
+    N_("Snapshot sync counters differ by more than 1"),
+    snapshot_print_fn
+  }
+};
+    
+void
+snapshot_handler (struct handler_param *param)
+{
+  char *sa = tildexpand (PARAM_STRING (param, 0));
+  char *sb = tildexpand (PARAM_STRING (param, 1));
+  char const *sel;
+  int rc = gdbm_latest_snapshot (sa, sb, &sel); 
+
+  if (rc >= 0 && rc < ARRAY_SIZE (snapshot_status_info))
+    {
+      fprintf (param->fp,
+	       "%s: %s.\n", 
+	       snapshot_status_info[rc].code,
+	       gettext (snapshot_status_info[rc].descr));
+      if (snapshot_status_info[rc].fn)
+	snapshot_status_info[rc].fn (param->fp, sa, sb);
+      if (rc == GDBM_SNAPSHOT_OK)
+	print_snapshot (sel, param->fp);
+    }
+  else
+    terror (_("unexpected error code: %d"), rc);
+}
+
 
 /* hash KEY - hash the key */
 void
@@ -1090,6 +1355,55 @@ debug_handler (struct handler_param *param)
   terror ("%s", _("compiled without debug support"));
 #endif
 }
+
+void
+shell_handler (struct handler_param *param)
+{
+  char *argv[4];
+  pid_t pid, rc;
+  int status;
+  
+  argv[0] = getenv ("$SHELL");
+  if (!argv[0])
+    argv[0] = "/bin/sh";
+  if (param->vararg)
+    {
+      argv[1] = "-c";
+      argv[2] = param->vararg->v.string;
+      argv[3] = NULL;
+    }
+  else
+    {
+      argv[1] = NULL;
+    }
+
+  pid = fork ();
+  if (pid == -1)
+    {
+      terror ("fork: %s", strerror (errno));
+      return;
+    }
+  if (pid == 0)
+    {
+      execv (argv[0], argv);
+      _exit (127);
+    }
+
+  rc = waitpid (pid, &status, 0);
+  if (rc == -1)
+    terror ("waitpid: %s\n", strerror (errno));
+  else if (!interactive ())
+    {
+      if (WIFEXITED (status))
+	{
+	  if (WEXITSTATUS (status) != 0)
+	    terror ("command failed with status %d", WEXITSTATUS (status));
+	}
+      else if (WIFSIGNALED (status))
+	terror ("command terminated on signal %d", WTERMSIG (status));
+    }
+}
+
 
 void
 source_handler (struct handler_param *param)
@@ -1269,6 +1583,32 @@ struct command command_tab[] = {
     FALSE,
     REPEAT_NEVER,
     N_("print current program status") },
+  { S(sync), T_CMD,
+    checkdb_begin, sync_handler, NULL,
+    { { NULL } },
+    FALSE,
+    REPEAT_NEVER,
+    N_("Synchronize the database with disk copy") },  
+  { S(upgrade), T_CMD,
+    checkdb_begin, upgrade_handler, NULL,
+    { { NULL } },
+    FALSE,
+    REPEAT_NEVER,
+    N_("Upgrade the database to extended format") },
+  { S(downgrade), T_CMD,
+    checkdb_begin, downgrade_handler, NULL,
+    { { NULL } },
+    FALSE,
+    REPEAT_NEVER,
+    N_("Downgrade the database to standard format") },    
+  { S(snapshot), T_CMD,
+    NULL, snapshot_handler, NULL,
+    { { "FILE", GDBM_ARG_STRING },
+      { "FILE", GDBM_ARG_STRING },
+      { NULL } },
+    FALSE,
+    REPEAT_NEVER,
+    N_("analyze two database snapshots") },
   { S(version), T_CMD,
     NULL, print_version_handler, NULL,
     { { NULL } },
@@ -1342,6 +1682,12 @@ struct command command_tab[] = {
     TRUE,
     REPEAT_NEVER,
     N_("query/set debug level") },
+  { S(shell), T_SHELL,
+    NULL, shell_handler, NULL,
+    { { NULL } },
+    TRUE,
+    REPEAT_NEVER,
+    N_("invoke the shell") },          
 #undef S
   { 0 }
 };
@@ -1355,9 +1701,9 @@ cmdcmp (const void *a, const void *b)
 }
 
 void
-sort_commands ()
+sort_commands (void)
 {
-  qsort (command_tab, sizeof (command_tab) / sizeof (command_tab[0]) - 1,
+  qsort (command_tab, ARRAY_SIZE (command_tab) - 1,
 	 sizeof (command_tab[0]), cmdcmp);
 }
 
@@ -1402,7 +1748,7 @@ int
 help_begin (struct handler_param *param GDBM_ARG_UNUSED, size_t *exp_count)
 {
   if (exp_count)
-    *exp_count = sizeof (command_tab) / sizeof (command_tab[0]) + 1;
+    *exp_count = ARRAY_SIZE (command_tab) + 1;
   return 0;
 }
 
@@ -1416,14 +1762,25 @@ help_handler (struct handler_param *param)
     {
       int i;
       int n;
-
+      int optoff;
+      
       n = fprintf (fp, " %s", cmd->name);
-
+      optoff = n;
+      
       for (i = 0; i < NARGS && cmd->args[i].name; i++)
-	n += fprintf (fp, " %s", gettext (cmd->args[i].name));
+	{
+	  if (n >= CMDCOLS)
+	    {
+	      fputc ('\n', fp);
+	      n = fprintf (fp, "%*.*s", optoff, optoff, "");
+	    }
+	  n += fprintf (fp, " %s", gettext (cmd->args[i].name));
+	} 
 
       if (n < CMDCOLS)
 	fprintf (fp, "%*.s", CMDCOLS-n, "");
+      else
+	fprintf (fp, "\n%*.*s", CMDCOLS, CMDCOLS, "");
       fprintf (fp, " %s", gettext (cmd->doc));
       fputc ('\n', fp);
     }
@@ -1438,7 +1795,8 @@ command_lookup (const char *str, struct locus *loc, struct command **pcmd)
   
   for (cmd = command_tab; state != fcom_abort && cmd->name; cmd++)
     {
-      if (memcmp (cmd->name, str, len < cmd->len ? len : cmd->len) == 0)
+      size_t n = len < cmd->len ? len : cmd->len;
+      if (memcmp (cmd->name, str, n) == 0 && str[n] == 0)
 	{
 	  switch (state)
 	    {
@@ -1500,6 +1858,12 @@ struct gdbm_option optab[] = {
   { 'r', "read-only",  NULL,       N_("open database in read-only mode") },
   { 's', "synchronize", NULL,      N_("synchronize to disk after each write") },
   { 'q', "quiet",      NULL,       N_("don't print initial banner") },
+  { 'd', "db-descriptor",
+    /* TRANSLATORS: File Descriptor. */
+    N_("FD"),
+    N_("open database at the given file descriptor") },
+  { 'x', "extended",   NULL,       N_("extended format (numsync)") },
+  {   0, "numsync",    NULL,       NULL, PARSEOPT_ALIAS },
 #if GDBMTOOL_DEBUG    
   { OPT_LEX_TRACE, "lex-trace", NULL, N_("enable lexical analyzer traces") },
   { OPT_GRAM_TRACE, "gram-trace", NULL, N_("enable grammar traces") },
@@ -1722,13 +2086,13 @@ struct handler_param param;
 size_t argmax;
 
 void
-param_free_argv (struct handler_param *param, int n)
+param_free_argv (struct handler_param *p, int n)
 {
   int i;
 
   for (i = 0; i < n; i++)
-    gdbmarg_destroy (&param->argv[i]);
-  param->argc = 0;
+    gdbmarg_destroy (&p->argv[i]);
+  p->argc = 0;
 }
 
 typedef struct gdbmarg *(*coerce_type_t) (struct gdbmarg *arg,
@@ -2035,6 +2399,15 @@ main (int argc, char *argv[])
        opt = parseopt_next ())
     switch (opt)
       {
+      case 'd':
+	file_descr = atoi (optarg);
+	if (file_descr <= 0)
+	  {
+	    terror (_("invalid file descriptor: %s"), optarg);
+	    exit (EXIT_USAGE);
+	  }
+	break;
+	
       case 'f':
 	source = optarg;
 	break;
@@ -2083,6 +2456,10 @@ main (int argc, char *argv[])
 	variable_set ("quiet", VART_BOOL, &bv);
 	break;
 
+      case 'x':
+	variable_set ("format", VART_STRING, "numsync");
+	break;
+	
       case OPT_LEX_TRACE:
 	lex_trace (1);
 	break;
@@ -2092,8 +2469,8 @@ main (int argc, char *argv[])
 	break;
 	
       default:
-	terror (_("unknown option; try `%s -h' for more info"),
-		progname);
+	terror (_("unknown option %c; try `%s -h' for more info"),
+		optopt, progname);
 	exit (EXIT_USAGE);
       }
   
